#version 430 core

#define PI 3.1415926538

in VertexData {
	vec3 position_world;
	vec3 normal_world;
	vec2 uv;
} vert;

layout (location = 0) out vec4 color;
layout (location = 1) out vec4 BloomColor;


uniform vec3 camera_world;

uniform vec3 materialCoefficients; // x = ambient, y = diffuse, z = specular 
uniform float specularAlpha;
uniform sampler2D diffuseTexture;

uniform float metalic;
uniform float roughness;

uniform struct DirectionalLight {
	vec3 color;
	vec3 direction;
} dirL;

uniform struct PointLight {
	vec3 color;
	vec3 position;
	vec3 attenuation;
} pointL;


vec3 fresnel(float HdotV, vec3 reflectivity){
	return reflectivity + (1.0 - reflectivity) * pow(1.0 - HdotV, 5.0);
}

float geometricAttenuation(float NdotH, float NdotV, float NdotL, float HdotV) {
	float term1 = (2*NdotH*NdotV)/(HdotV);
	float term2 = (2*NdotH*NdotL)/(HdotV);
	float minTerm = min(term1, term2);
	return min(1.0, minTerm);
}

float distributionGGX(vec3 N, vec3 H, float a)
{
    float a2     = a*a;
    float NdotH  = max(dot(N, H), 0.0);
    float NdotH2 = NdotH*NdotH;
	
    float nom    = a2;
    float denom  = (NdotH2 * (a2 - 1.0) + 1.0);
    denom        = PI * denom * denom;
	
    return nom / denom;
}


void main() {	
	vec3 N = normalize(vert.normal_world);
	vec3 V = normalize(camera_world - vert.position_world);
	
	vec3 texColor = texture(diffuseTexture, vert.uv).rgb;
	color = vec4(texColor * materialCoefficients.x, 1); // ambient

	vec3 reflectivity = mix(vec3(0.03), texColor, metalic);
	
	// point light
	vec3 L = normalize(pointL.position - vert.position_world);
	vec3 H = normalize(V + L);
	float lightDistance = length(pointL.position - vert.position_world);
	float attenuation = 1.0 / (lightDistance * lightDistance);
	vec3 radiance = attenuation * pointL.color;

	float HdotV = max(dot(H,V), 0.0000001);
	float NdotH = max(dot(N,H), 0.0000001);
	float NdotL = max(dot(N,L), 0.0000001);
	float NdotV = max(dot(N,V), 0.0000001);

	float G = geometricAttenuation(NdotH, NdotV, NdotL, HdotV);
	float D = distributionGGX(N, H, roughness);
	vec3 F = fresnel(HdotV, reflectivity);

	vec3 specular = D*G*F;
	specular /= 4.0 * NdotV * NdotL;

	vec3 kD = vec3(1.0) - F; // F represents the specular amount and specular + diffuse <= 1.0 because of energy conservation
	kD *= 1.0 - metalic; // more metalic, less diffuse

	color.rgb += (kD * texColor + specular) * NdotL * radiance;

	// directional light
	L = normalize(dirL.direction - vert.position_world);
	H = normalize(V + L);
	radiance = dirL.color;

	HdotV = max(dot(H,V), 0.0000001);
	NdotH = max(dot(N,H), 0.0000001);
	NdotL = max(dot(N,L), 0.0000001);
	NdotV = max(dot(N,V), 0.0000001);

	G = geometricAttenuation(NdotH, NdotV, NdotL, HdotV);
	D = distributionGGX(N, H, roughness);
	F = fresnel(HdotV, reflectivity);

	specular = D*G*F;
	specular /= 4.0 * NdotV * NdotL;

	kD = vec3(1.0) - F; // F represents the specular amount and specular + diffuse <= 1.0 because of energy conservation
	kD *= 1.0 - metalic; // more metalic, less diffuse

	color.rgb += (kD * texColor + specular) * NdotL * radiance;

	BloomColor = vec4(0.0f,0.0f,0.0f,0.0f);

}

